#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <time.h>
#include <unistd.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <net/if_arp.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <libubox/list.h>
#include <libubox/uloop.h>
#include <libubus.h>
#include <uci.h>

#include <easy/easy.h>

#include "debug.h"
#include "host_config.h"
#include "topologyd.h"
#include "mdns_avahi.h"


void node_destroy(struct node *n)
{
	clean_mdns_services(n->device_ipv4);
	uloop_timeout_cancel(&n->validity_timer);
	free(n);
}

struct node *node_create(uint8_t *hwaddr)
{
	struct node *n;

	n = calloc(1, sizeof(struct node));
	if (!n) {
		warn("OOM: Node alloc failed!\n");
		return NULL;
	}

	memcpy(n->hwaddr, hwaddr, 6);

	return n;
}

// TODO: convert into macro usable by ipv4 and 6?
struct node *node_lookup_ipv4(struct hlist_head *table, uint32_t ipv4_addr)
{
	struct node *n = NULL;
	int i;

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&table[i]))
			continue;

		hlist_for_each_entry(n, &table[i], hlist) {
			int j;

			dbg("Node: " MACFMT ", ipv4_intf_nbr = %d\n",
					MAC2STR(n->hwaddr), n->ipv4_intf_nbr);

			for (j = 0; j < n->ipv4_intf_nbr; j++) {
				struct device_ipv4_t dev;

				dev = n->device_ipv4[j];

				dbg("1 -- 1st octet = %d 2nd = %d 3rd = %d 4th = %d\n",
						dev.ipv4_addr[0], dev.ipv4_addr[1],
						dev.ipv4_addr[2], dev.ipv4_addr[3]);
				dbg("2 -- 1st octet = %d 2nd = %d 3rd = %d 4th = %d\n",
						ipv4_addr & 0xFF, (ipv4_addr >> 8) & 0xFF,
						(ipv4_addr >> 16) & 0xFF, (ipv4_addr >> 24) & 0xFF);

				if (!(dev.ipv4_addr[0] ^ (ipv4_addr & 0xFF)) &&
						!(dev.ipv4_addr[1] ^ ((ipv4_addr >> 8) & 0xFF)) &&
						!(dev.ipv4_addr[2] ^ ((ipv4_addr >> 16) & 0xFF)) &&
						!(dev.ipv4_addr[3] ^ ((ipv4_addr >> 24) & 0xFF))) {
					dbg("Node %d.%d.%d.%d was found in htable\n",
							dev.ipv4_addr[0], dev.ipv4_addr[1],
							dev.ipv4_addr[2], dev.ipv4_addr[3]);
					return n;
				}
			}
		}
	}

	return NULL;
}

// TODO: convert into macro usable by ipv4 and 6?
struct node *node_lookup_ipv6(struct hlist_head *table, const uint8_t *ipv6_addr)
{
	struct node *n = NULL;
	int i;

	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&table[i]))
			continue;

		hlist_for_each_entry(n, &table[i], hlist) {
			int j;

			for (j = 0; j < n->ipv6_intf_nbr; j++) {
				struct device_ipv6_t dev;
				int k;
				bool match = true;

				dev = n->device_ipv6[j];
				for (k = 0; k < 16; k++) {
					if (dev.ipv6_addr[k] ^ ipv6_addr[k])
						match = false;
				}

				if (!match)
					break;

				return n;
			}
		}
	}

	return NULL;
}

struct node *node_lookup(struct hlist_head *table, uint8_t *hwaddr)
{
	int hidx = node_hash(hwaddr);
	struct node *n = NULL;

	if (hwaddr_is_zero(hwaddr))
		return NULL;

	hlist_for_each_entry(n, &table[hidx], hlist) {
		if (!memcmp(hwaddr, n->hwaddr, 6))
			return n;
	}

	return NULL;
}

int node_del(struct hlist_head *table, uint8_t *hwaddr)
{
	int hidx = node_hash(hwaddr);
	struct node *n;

	n = node_lookup(table, hwaddr);
	if (!n)
		return -1;

	hlist_del(&n->hlist, &table[hidx]);
	dbg("Node " MACFMT " removed from topology\n", MAC2STR(n->hwaddr));

	node_destroy(n);
	return 0;
}

struct node *node_add(struct hlist_head *table, uint8_t *hwaddr)
{
	struct node *n;

	n = node_lookup(table, hwaddr);
	if (n) {
		dbg("Node " MACFMT " already in topology\n", MAC2STR(hwaddr));
		return n;
	}

	n = node_create(hwaddr);
	if (n) {
		int hidx = node_hash(hwaddr);

		hlist_add_head(&n->hlist, &table[hidx]);
		dbg("Node " MACFMT " added to topology\n", MAC2STR(hwaddr));
	}

	return n;
}

void node_print_all(struct hlist_head *table)
{
	struct node *n;
	int i;

	dbg("Nodes:\n");
	for (i = 0; i < NODE_HTABLE_SIZE; i++) {
		if (hlist_empty(&table[i]))
			continue;

		hlist_for_each_entry(n, &table[i], hlist) {
			dbg("Node: " MACFMT "\n", MAC2STR(n->hwaddr));
		}
	}
}
